from sqlalchemy import create_engine
from sqlalchemy.orm import declarative_base, sessionmaker

SQLALCHEMY_DATABASE_URL = "sqlite:///./models.db"

engine = create_engine(
    SQLALCHEMY_DATABASE_URL, connect_args={"check_same_thread": False}
)
SessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=engine)
Base = declarative_base()



import os
import pytest
from unittest.mock import Mock, patch
from dpai import (
    run_inference,
    start_app,
)
from dpai.model_manager import InferenceException


dir_path = os.path.dirname(os.path.realpath(__file__))


class TestUtils:
    @pytest.fixture(autouse=True)
    def init_model(self):
        self._model_name = "mock_model_name"
        self._inference_path = "mock_inference_path"
        self._model_path = "mock_model_path"
        self._specs = {}
        self._port = 9000
        self._model_manager = Mock()

    @pytest.mark.parametrize(
        "data, expected", [({"data": "mock_data"}, "prediction_result")]
    )
    @patch("importlib.util.spec_from_file_location")
    @patch("importlib.util.module_from_spec")
    @patch("dpai.utils.load")
    def test_run_inference(
        self,
        mock_load,
        mock_module_from_spec,
        mock_spec_from_file_location,
        data,
        expected,
    ):
        mock_spec_from_file_location.return_value = Mock()
        mock_inference_file = Mock()
        mock_module_from_spec.return_value = mock_inference_file
        mock_load.return_value = {"model": Mock()}

        # TEST 1: error transforming input
        mock_inference_file.input_fn.side_effect = Exception("Input Function Error")

        with pytest.raises(InferenceException) as exc_info:
            result = run_inference(data, self._model_path, self._inference_path)
        mock_load.assert_called_once_with(self._model_path)
        mock_inference_file.input_fn.assert_called_once_with(data)
        mock_inference_file.predict_fn.assert_not_called()
        assert "Failed to load data" in str(exc_info.value)

        # TEST 2: error during inference
        mock_inference_file.input_fn.side_effect = None  # Reset side effect
        mock_inference_file.input_fn.return_value = Mock()
        mock_inference_file.predict_fn.side_effect = Exception("Predict Function Error")

        with pytest.raises(InferenceException) as exc_info:
            result = run_inference(data, self._model_path, self._inference_path)
        mock_inference_file.predict_fn.assert_called_once_with(
            mock_inference_file.input_fn.return_value, mock_load.return_value["model"]
        )
        assert "Failed to run inference" in str(exc_info.value)

        # Test 3: successful run
        mock_inference_file.predict_fn.side_effect = None
        mock_inference_file.predict_fn.return_value = "prediction_result"

        result = run_inference(data, self._model_path, self._inference_path)
        assert result == expected

    @patch("dpai.utils.uvicorn.run")
    @patch("dpai.model_manager.ModelManager")
    def test_start_app(self, mock_model_manager, mock_uvicorn_run):
        start_app(self._port, mock_model_manager.return_value)

        # Assertions
        mock_model_manager.return_value.load_models.assert_called_once()
        mock_uvicorn_run.assert_called_once_with(
            "fastapi_main:app", port=self._port, reload=True
        )



__version__ = "0.1.0"

from .fastapi_main import app, Input
from .main import main
from .model_manager import ModelManager
from .utils import (
    run_inference,
    start_app,
)

__all__ = [
    "__version__",
    "main",
    "app",
    "Input",
    "ModelManager",
    "ModelException",
    "run_inference",
    "start_app",
]



from typing import List
from contextlib import asynccontextmanager

from fastapi import FastAPI, status, HTTPException, Depends
from pydantic import BaseModel
from sqlalchemy.orm import Session

from .utils import run_inference
from .model_db.crud import get_ml_model
from .model_db.db_setup import Base, SessionLocal, engine


class Input(BaseModel):
    data: List[str]


@asynccontextmanager
async def lifespan(app: FastAPI):
    # startup
    Base.metadata.create_all(bind=engine)
    yield
    # shutdown
    Base.metadata.drop_all(bind=engine)


app = FastAPI(lifespan=lifespan)


# Dependency to get the database session, ignored during test
def get_db():  # pragma: no cover
    database = SessionLocal()
    try:
        yield database
    finally:
        database.close()


@app.get("/")
async def root():
    return {"message": "Hello World"}


@app.post("/{model_name}/predict")
async def predict(model_name: str, request_body: Input, db: Session = Depends(get_db)):
    model = get_ml_model(db, model_name)
    if model is None:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail="Invalid model name, this model is not in registry",
        )

    model_path, inference_path = model.model_path or "", model.inference_path or ""

    try:
        result = run_inference(request_body.data, model_path, inference_path)
    except Exception as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f"Error running inference, check your inference script. Error: {str(e)}",
        )
    if isinstance(result, str):
        return {"response": result}
    else:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail="Response format is not string, check your inference file's predict_fn again",
        )



from sqlalchemy.orm import Session

from .models import MLModel
from typing import List


def get_ml_model(db: Session, model_name: str) -> MLModel | None:
    return db.query(MLModel).filter(MLModel.name == model_name).first()


def get_all_models(db: Session) -> List[MLModel]:
    return db.query(MLModel).all()


def create_ml_model(
    db: Session, model_name: str, model_path: str, inference_path: str
) -> None:
    db_model = MLModel(
        name=model_name,
        model_path=model_path,
        inference_path=inference_path,
    )
    db.add(db_model)
    db.commit()
    db.refresh(db_model)


# def get_users(db: Session, skip: int = 0, limit: int = 100):
#     return db.query(models.User).offset(skip).limit(limit).all()

# def get_items(db: Session, skip: int = 0, limit: int = 100):
#     return db.query(models.Item).offset(skip).limit(limit).all()


# def create_user_item(db: Session, item: schemas.ItemCreate, user_id: int):
#     db_item = models.Item(**item.dict(), owner_id=user_id)
#     db.add(db_item)
#     db.commit()
#     db.refresh(db_item)
#     return db_item



from sqlalchemy import create_engine, StaticPool
from sqlalchemy.orm import sessionmaker
from fastapi.testclient import TestClient
from unittest.mock import patch, Mock
import pytest

from dpai.fastapi_main import app, Base, get_db

# Setup the TestClient
client = TestClient(app)

# Setup the in-memory SQLite database for testing
DATABASE_URL = "sqlite:///:memory:"
engine = create_engine(
    DATABASE_URL,
    connect_args={
        "check_same_thread": False,
    },
    poolclass=StaticPool,
)
TestingSessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=engine)


# Dependency to override the get_db dependency in the main app
def override_get_db():
    database = TestingSessionLocal()
    yield database
    database.close()


app.dependency_overrides[get_db] = override_get_db


@pytest.fixture(scope="module")
def test_app():
    with TestClient(app) as client:
        yield client


def test_root(test_app):
    response = test_app.get("/")
    assert response.status_code == 200
    assert response.json() == {"message": "Hello World"}


@patch("dpai.fastapi_main.run_inference")
@patch("dpai.fastapi_main.get_ml_model")
def test_predict_success(mock_get_ml_model, mock_run_inference, test_app):
    # Mock get_ml_model to return model paths
    mock_get_ml_model.return_value = Mock(
        model_path="mock_path", inference_path="mock_path"
    )

    # Mock run_inference to return a string response
    mock_run_inference.return_value = "Prediction result"

    response = test_app.post("/mock_model/predict", json={"data": ["example_data"]})

    mock_run_inference.assert_called_once_with(
        ["example_data"], "mock_path", "mock_path"
    )
    assert response.status_code == 200
    assert response.json() == {"response": "Prediction result"}


@patch("dpai.fastapi_main.run_inference")
@patch("dpai.fastapi_main.get_ml_model")
def test_predict_invalid_model(mock_get_ml_model, mock_run_inference, test_app):
    # Failed model retrieval
    mock_get_ml_model.return_value = None
    response = test_app.post("/mock_model/predict", json={"data": ["example_data"]})

    mock_run_inference.assert_not_called()
    assert response.status_code == 400
    assert response.json() == {
        "detail": "Invalid model name, this model is not in registry"
    }


@patch("dpai.fastapi_main.run_inference")
@patch("dpai.fastapi_main.get_ml_model")
def test_predict_invalid_inference_result(
    mock_get_ml_model, mock_run_inference, test_app
):
    # inference result is problematic
    mock_get_ml_model.return_value = Mock(
        model_path="mock_path", inference_path="mock_path"
    )
    mock_run_inference.return_value = {"result": "not_a_string"}

    response = test_app.post("/mock_model/predict", json={"data": ["example_data"]})

    mock_run_inference.assert_called_once_with(
        ["example_data"], "mock_path", "mock_path"
    )
    assert response.status_code == 500
    assert response.json() == {
        "detail": "Response format is not string, check your inference file's predict_fn again"
    }


@patch("dpai.fastapi_main.run_inference")
@patch("dpai.fastapi_main.get_ml_model")
def test_predict_inference_error(mock_get_ml_model, mock_run_inference, test_app):
    # inference result is problematic
    mock_get_ml_model.return_value = Mock(
        model_path="mock_path", inference_path="mock_path"
    )
    mock_run_inference.side_effect = Exception("error message")
    response = test_app.post("/mock_model/predict", json={"data": ["example_data"]})

    mock_run_inference.assert_called_once_with(
        ["example_data"], "mock_path", "mock_path"
    )
    assert response.status_code == 500
    assert response.json() == {
        "detail": "Error running inference, check your inference script. Error: error message"
    }


def setup() -> None:
    # Create the tables in the test database
    Base.metadata.create_all(bind=engine)


def teardown() -> None:
    # Drop the tables in the test database
    Base.metadata.drop_all(bind=engine)



from typing import Generator
from sqlalchemy import create_engine, StaticPool
from sqlalchemy.orm import sessionmaker, Session

from dpai.model_db.models import Base, MLModel
from dpai.model_db.crud import get_ml_model, get_all_models, create_ml_model

import pytest


DATABASE_URL = "sqlite:///:memory:"
engine = create_engine(
    DATABASE_URL,
    connect_args={
        "check_same_thread": False,
    },
    poolclass=StaticPool,
)
TestingSessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=engine)


def override_get_db():
    database = TestingSessionLocal()
    yield database
    database.close()


@pytest.fixture
def db_item():
    return MLModel(id=1, name="test_model", model_path="", inference_path="")


@pytest.fixture
def db_session(db_item) -> Generator[Session, None, None]:
    # Create an in-memory SQLite database for testing
    Base.metadata.create_all(bind=engine)
    db_session = TestingSessionLocal()

    db_session.add(db_item)
    db_session.commit()

    yield db_session

    # Tear down the database after testing
    db_session.close()
    Base.metadata.drop_all(bind=engine)


@pytest.mark.parametrize("model_name", [("test_model")])
def test_get_ml_model_existing_model(db_session, model_name, db_item):
    # Create a mock MLModel object in the database
    result = get_ml_model(db_session, model_name)

    # Assert that the returned MLModel object matches the one created
    assert result.id == db_item.id
    assert result.name == db_item.name
    assert result.model_path == db_item.model_path
    assert result.inference_path == db_item.inference_path


@pytest.mark.parametrize("model_name", [("nonexistent_model")])
def test_get_ml_model_nonexistent_model(db_session, model_name):
    # Call the function under test with a model name that doesn't exist
    result = get_ml_model(db_session, model_name)

    assert result is None


def test_get_all_models(db_session, db_item):
    result = get_all_models(db_session)
    assert len(result) == 1
    assert result[0].id == db_item.id
    assert result[0].name == db_item.name
    assert result[0].model_path == db_item.model_path
    assert result[0].inference_path == db_item.inference_path


def test_create_ml_model(db_session):
    # Call the function under test to create a new MLModel
    model_name = "new_model"
    model_path = "/path/to/model"
    inference_path = "/path/to/inference"
    create_ml_model(db_session, model_name, model_path, inference_path)

    # Query the database to get the created MLModel
    db_model = db_session.query(MLModel).filter_by(name=model_name).first()

    # Assert that the created MLModel matches the expected values
    assert db_model is not None
    assert db_model.name == model_name
    assert db_model.model_path == model_path
    assert db_model.inference_path == inference_path



from unittest.mock import patch, Mock
import pytest
from dpai.model_manager import ModelManager, ModelException


@pytest.fixture
def mock_session():
    with patch("dpai.model_manager.SessionLocal") as mock_session:
        yield mock_session.return_value


@patch("dpai.model_manager.create_ml_model")
@patch("dpai.model_manager.get_ml_model")
def test_register_model_existing_model(
    mock_get_ml_model, mock_create_ml_model, mock_session
):
    # Mock get_ml_model to return a non-None value, indicating an existing model
    mock_get_ml_model.return_value = Mock()

    model_manager = ModelManager()

    # Assert that ModelException is raised
    with pytest.raises(ModelException) as exc_info:
        model_manager.register_model("existing_model", "model_path", "inference_path")

    mock_get_ml_model.assert_called_once_with(mock_session, "existing_model")
    mock_create_ml_model.assert_not_called()
    assert "Model with the same already register" in str(exc_info.value)


@patch("dpai.model_manager.create_ml_model")
@patch("dpai.model_manager.get_ml_model")
def test_register_model_new_model(
    mock_get_ml_model, mock_create_ml_model, mock_session
):
    # Mock get_ml_model to return None, indicating a new model
    mock_get_ml_model.return_value = None

    model_manager = ModelManager()

    # Call the function under test
    model_manager.register_model("new_model", "model_path", "inference_path")

    mock_get_ml_model.assert_called_once_with(mock_session, "new_model")
    mock_create_ml_model.assert_called_once_with(
        mock_session, "new_model", "model_path", "inference_path"
    )


@patch("dpai.model_manager.get_all_models")
def test_load_models(mock_get_all_models, mock_session):
    # Mock get_all_models to return some mock models
    mock_model_1 = Mock()
    mock_model_2 = Mock()
    mock_get_all_models.return_value = [mock_model_1, mock_model_2]

    model_manager = ModelManager()

    # Call the function under test
    model_manager.load_models()

    # Assertions
    assert len(model_manager._ModelManager__models) == 2  # Check if models are loaded
    mock_get_all_models.assert_called_once_with(
        mock_session
    )  # Check if get_all_models is called


@patch("dpai.model_manager.get_all_models")
def test_load_models_with_exception(mock_get_all_models, mock_session):
    # Mock get_all_models to return some mock models
    model_manager = ModelManager()

    mock_get_all_models.side_effect = Exception()
    with pytest.raises(ModelException) as exc_info:
        _ = model_manager.load_models()

    # Assertions
    assert "Cannot get models from db" in str(exc_info.value)
    mock_get_all_models.assert_called_once_with(
        mock_session
    )  # Check if get_all_models is called



from sqlalchemy import Column, Integer, String

from .db_setup import Base


class MLModel(Base):
    __tablename__ = "models"

    id = Column(Integer, primary_key=True)
    name = Column(String, unique=True, index=True)
    model_path = Column(String, nullable=False)
    inference_path = Column(String, nullable=False)



import setuptools


package_data = {"dpai": ["models/*", "model_files/*", "static/*", "templates/*"]}

setuptools.setup(
    packages=setuptools.find_namespace_packages("src"),
    package_dir={"": "src"},
    package_data=package_data,
)



import argparse
import os

from .utils import start_app
from .model_manager import ModelManager

# from setups import logger


def main():
    parser = argparse.ArgumentParser("dpai parser")
    subparser = parser.add_subparsers(dest="func")

    parser_register = subparser.add_parser("register", help="Register model")
    parser_register.add_argument("-n", "--name", type=str, required=True)
    parser_register.add_argument(
        "-m", "--model_path", required=True, type=os.path.abspath
    )
    parser_register.add_argument(
        "-i", "--inference_path", required=True, type=os.path.abspath
    )

    parser_serve = subparser.add_parser("serve", help="Serve model")
    parser_serve.add_argument("--port", type=int, default=9000)
    args = parser.parse_args()

    model_manager = ModelManager()

    if args.func == "register":
        # logger.info("Register")
        model_manager.register_model(args.name, args.model_path, args.inference_path)
        model_manager.load_models()

    elif args.func == "serve":
        # logger.info("serve")
        start_app(args.port, model_manager)
    else:
        raise NotImplementedError



import importlib.util as util
from typing import Any
from joblib import load

import uvicorn
from .model_manager import ModelManager, InferenceException


def run_inference(data: Any, model_path: str, inference_path: str) -> Any:
    """
    TODO: cannot load model every time, must either cache it or store it in local
    """
    spec = util.spec_from_file_location("inference", inference_path)
    assert spec is not None
    assert spec.loader is not None
    inference_file = util.module_from_spec(spec)
    spec.loader.exec_module(inference_file)

    model = load(model_path)["model"]
    try:
        inputs = inference_file.input_fn(data)
    except Exception as e:
        raise InferenceException("Failed to load data, causes:", e)

    try:
        outputs = inference_file.predict_fn(inputs, model)
        return outputs
    except Exception as e:
        raise InferenceException("Failed to run inference, causes:", e)


def start_app(port: int, model_manager: ModelManager) -> None:
    """Start the application"""
    # [TODO] save models / models' metadata to cache
    model_manager.load_models()
    uvicorn.run("fastapi_main:app", port=port, reload=True)



from .model_db.crud import create_ml_model, get_ml_model, get_all_models
from .model_db.db_setup import Base, SessionLocal, engine

# Spin up DB
Base.metadata.create_all(bind=engine)


class InferenceException(Exception):
    pass


class ModelException(Exception):
    pass


class ModelManager:
    """
    Model Manager class for loading, managing, and interacting with models
    class that instantiates and manages model objects.
    """

    def __init__(self):
        self.__models = []

    def register_model(
        self, model_name: str, model_path: str, inference_path: str
    ) -> None:
        """save the model metadata into model registry"""
        db = SessionLocal()
        if get_ml_model(db, model_name):
            raise ModelException("Model with the same already register")
        else:
            create_ml_model(db, model_name, model_path, inference_path)
            # update_model_list_in_config(MODEL_CONFIG_PATH, [model_name])
        db.close()

    def load_models(self):
        """Load models from configuration."""
        db = SessionLocal()
        try:
            self.__models = []  # resetting the model list to empty
            models = {}
            for model in get_all_models(db):
                self.__models.append(model)
                models[model.name] = {
                    "model_path": model.model_path,
                    "inference_path": model.inference_path,
                }
        except Exception as e:
            raise ModelException("Cannot get models from db", e)
        finally:
            db.close()

    # def get_model_metadata(self, model_name: str) -> tuple[str, str]:
    #     """output model metadata associated with the given model name"""
    #     repo = load_model_repo_from_yaml(MODEL_REPO_PATH)
    #     if model_name in repo:
    #         return repo[model_name]["model_path"], repo[model_name]["inference_path"]
    #     else:
    #         print(f"{model_name} not found in current model registry")
    #     return "", ""

    # @classmethod
    # def get_model_metadata_2(cls, qualified_name: str) -> Optional[Dict[str, Any]]:
    #     """Get a model metadata by qualified name."""
    #     # searching the list of model objects to find the one with the right qualified name
    #     model_objects = [
    #         model for model in cls.__models if model.qualified_name == qualified_name
    #     ]

    #     if len(model_objects) == 0:
    #         return None
    #     else:
    #         model_object = model_objects[0]
    #         return {
    #             "display_name": model_object.display_name,
    #             "qualified_name": model_object.qualified_name,
    #             "description": model_object.description,
    #             "major_version": model_object.major_version,
    #             "minor_version": model_object.minor_version,
    #             "input_schema": model_object.input_schema.to_json_schema(),
    #             "output_schema": model_object.output_schema.to_json_schema(),
    #         }

    # def get_models(self):
    #     return self.__models
    # def get_model(self, model_name: str) -> Optional[Model]:
    #     """Get a model object by qualified name."""
    #     # searching the list of model objects to find the one with the right qualified name
    #     model_objects = [
    #         model for model in self.__models if model.model_name == model_name
    #     ]

    #     if len(model_objects) == 0:
    #         return None
    #     else:
    #         return model_objects[0]
